///////////////////////////////////////////////////////////////////////
//
// David Cheung
// U49303740
// CS 540: AI
// HW 2
// Feb 24, 1999
//////////////////////////////////////////////////////////////////////
// tree.cpp: implementation of the tree class.
//
//////////////////////////////////////////////////////////////////////

#include "tree.h"
#include "node.h"
#include <stdio.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


tree::tree()
{
	root=NULL;
	curptr=NULL;
	numnode=0;
	printmode='d';
	int i=0;
	for(i=0;i<362880;i++)
		hashtable[i]=false;
}

tree::~tree()
{

}

void tree::addnode(node *anode, char type)
{
	//the use of addnode is to add
	// a node either as the child or 
	// sibling of the node. if type is "c"
	// then it means adding the new node
	// as a CHILD of the current node.
	// if the type is "s" then it means adding
	// the new node as a sibling of that node.

	// implementation problem:
	// what if the tree has 0 node.
	// what if the curptr's node already has a 
	// child and u are adding a child to it?
	// what if the curptr's node already has a 
	// sibling and u are adding a sibling to it?


	// it seems like easier to implement add sibling
	// first because once that's set, if the curptr
	// has a child and u need to add a child, then
	// just addsibling to that child.
	
	// in anycase, numnode will only increment if node
	// is added via here, not addchild or addsibling.
	if(numnode==0)
	{
		tree::root=anode;
		tree::curptr=anode;
		anode->depth=0;
	}
	else
	{
		if(type=='c') addchild(anode);
		else addsibling(curptr,anode);
		anode->depth=anode->parent->depth +1;
	}
	numnode++;


}

void tree::addchild(node *anode)
{

	if(!curptr->child)  // do this only if there's no child...
	{
		anode->parent=curptr;
		curptr->child=anode;
	}
	else
	{
		// adding node to the child's sibling.
		// i realized i can't just addsibling to this
		// because curptr is not pointing at the right place.
		// yet i don't want to change the curptr!
		// maybe if i add another parameter to addsibling....
		addsibling(curptr->child, anode);
	}

}

void tree::addsibling(node* from_ptr, node* anode)
{
	if(!from_ptr->sister)
	{
		anode->parent=from_ptr->parent;
		from_ptr->sister= anode;
	}
	else tree::addsibling(from_ptr->sister, anode);

};

void tree::printpath()
{
	// for dfs, printmode = 'd'
	// for bfs, printmode = 'b'
	// for path, printmode = 'p'
	printmode='p';
	node *tmpptr=curptr;
	if(tmpptr)
	{
		recurseprint(tmpptr);

	}
	else return;
}

void tree::recurseprint(node *tmpptr)
{
	if(printmode=='b')
	{
		// perform depth first search printing
		tmpptr->printdata();
		if(tmpptr->child) recurseprint(tmpptr->child);
		if(tmpptr->sister) recurseprint(tmpptr->sister);
	}
	else
	{
		if(printmode=='p')
		{
			// perform path printing
			if(tmpptr->parent) recurseprint(tmpptr->parent);
			tmpptr->printdata();
		}
	}

  
}


void tree::printbfs()
{
	// for dfs, printmode = 'd'
	// for bfs, printmode = 'b'
	printmode='b';
	node *tmpptr=root;
	if(tmpptr)
	{
		recurseprint(tmpptr);
	}
	else return;

}

bool tree::recurseismember(node *tmpptr,node *tnode)
{
	if(datamatch(tmpptr, tnode))
		return true;
	else
	{
		bool test=false;
		if(tmpptr->child)	
			test = recurseismember(tmpptr->child, tnode);
		if(tmpptr->sister)
			test =  test || recurseismember(tmpptr->sister,tnode);
		return test;
	}
	//2/23/99: this function is tested and works
	// accordingly...
}


bool tree::ismember(node *tnode)
{
	node *tmpptr;
	tmpptr=tree::root;
	if(numnode >= 1)
	{
		return	recurseismember(tmpptr,tnode);
	}
	return false;
	//2/23/99: THis function is tested and works
}


bool tree::querydata(node *tnode)
{
	//tested...this function works!
	node *tmpptr;
	tmpptr=tree::curptr;
	if(numnode >= 1)
	{
		while(tmpptr)
		{
			if (datamatch(tmpptr, tnode))
				return true;
			else
				tmpptr=tmpptr->parent;
		}
	}

	return false;
}

bool tree::datamatch(node *a, node *b)
{

	int i = 0;
	for (i=0;i<9;i++)
	{
		if (!(a->data[i]==b->data[i]))
		{
			return false;
		}
	}
	return true;
}


void tree::printpathfile(FILE *ofp)
{
	node *tmpptr=curptr;
	if(tmpptr)
	{
		recursefileprint(tmpptr,ofp);

	}
	else return;

}

void tree::recursefileprint(node *tmpptr,FILE *ofp)
{

	if(tmpptr->parent) recursefileprint(tmpptr->parent,ofp);
	tmpptr->printfile(ofp);

}

int tree::conversionfactor(int array[])
{
  int i=0,j=0,value=8,factor=0,values[9];
  bool done = false;
  
//  for (i=0;i<3;i++)
//  for (j=0;j<3;j++) 
//      values[i*3+j] = board[i][j]-1;
  
  for(i=0;i<9;i++)
	  values[i]=array[i] -1;
  i=0;

  do {
    
    if (values[i] == value) {
      done = true;
      for (j=0;j<9;j++)
	factor += values[j]*factorial(8-j);
    }
    else {
      value--;
      for (j=i+1;j<9;j++)
	if (values[j] > values[i])
	  values[j]--;
    }
    i++;
  }
  while ((i < 9) && (!done));
  
  return factor;

}

int tree::factorial(int num)
{

  int value=1,i;
  
  for (i=num;i>1;i--)
    value *= i;
  
  return value;

}
